/****************************************************************************
Copyright 2010 Bilkent University

Licensed under the Apache License, Version 2.0 (the "License"); 
you may not use this file except in compliance with the License. 
You may obtain a copy of the License at 
http://www.apache.org/licenses/LICENSE-2.0 

Unless required by applicable law or agreed to in writing, software 
distributed under the License is distributed on an "AS IS" BASIS, 
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
See the License for the specific language governing permissions and 
limitations under the License. 
****************************************************************************/
#ifndef _RENG_GPUQUERY_H_
#define _RENG_GPUQUERY_H_

#include "REng/Prerequisites.h"

#include "REng/Defines.h"
#include "REng/GPU/GPUResource.h"

// The stuff below are not available with OpenGL ES 2.0 configurations
#if RENG_GL_PLATFORM != RENG_GL_PLATFORM_ES

namespace REng{

	/*!
	 *  Asynchronous queries provide a mechanism to return information about the processing of
	 *  a sequence of GL commands.
	 *  @note Asynchronous queries are not bindable, yet are started / ended.
	 *  @author Adil Yalcin
	 */ 
	class RENGAPI GPUQuery : public GPUResource{
	public:

		//! @brief The destructor automatically deletes any hardware resource assigned
		virtual ~GPUQuery();

		//! @return True if this asynchronous query is currently active / data is tracked.
		bool isActive() const;

		//! @return True if the asynchronous query of given type is currently active / data is tracked.
		static bool isActive(QueryTarget target);

		//! Checks whether there may be an indeterminate delay before the above query returns
		bool resultAvailable() const;
		//! Returns the query object result value as a signed integer
		void getResultInt(int* p) const;
		//! Returns the query object result value as an unsigned integer
		void getResultUInt(uint* p) const;
		//! @note For use by any-sample-passed occlusion query
		void getResultBool(bool* p) const;

		virtual void begin() = 0;
		virtual void end() = 0;

	protected:
		GPUQuery();

		//! The shared/generic object-specific function that wraps glBeginQuery interface
		void beginQuery(QueryTarget target);
		//! The shared/generic object-specific function that wraps glEndQuery interface
		void endQuery(QueryTarget target);

	private:
		void createResource();
		void destroyResource();

		//! True if the query is active
		//! @note A query can only be active for a single target at a time
		bool mIsActive;
		
		//! The active transform feedback query, 0 if not active
		static GLuint ActiveQueryPrimsGenerated;
		static GLuint ActiveQueryTransFeedback;
		static GLuint ActiveQueryOcclusion; // same for samples_passed and any_samples_passed
		static GLuint ActiveQueryTime;
		static void setActiveQueryObject(QueryTarget target, GLuint id);
	};

	/**
	 *  Primitive queries are used
	 *  - to track the number of primitives generated by the GL or
	 *  - to track the number of primitives written to transform feedback buffers
	 *  @note These two set of queries can be used together to determine if all primitives 
	 *        have been written to the bound feedback buffers.
	 *  @note This class in an intermediate, no-op class
	 *  @author Adil Yalcin
	 */
	class RENGAPI GPUQueryPrim : public GPUQuery {
	public:
		virtual ~GPUQueryPrim(){}
	};

	/**
	 *  The primitives-generated count is incremented every time a primitive reaches
	 *  the transform feedback stage,whether or not transform feedback is active.
	 *  
	 *  This counter counts the number of primitives emitted by a geometry shader,
	 *  if active, possibly further tessellated into separate primitives during the 
	 *  transform-feedback stage, if enabled.
	 *  @author Adil Yalcin
	 */
	class RENGAPI GPUQueryPrimGenerated : public GPUQueryPrim {
	public:
		GPUQueryPrimGenerated();

		//! Primitives-generated count maintained by the GL is set to zero.
		void begin();
		void end();
	};

	/**
	 *  The transform-feedback-primitives-written count is incremented every time a primitive
	 *  is recorded into a buffer object.
	 *
	 *  If transform feedback is not active, this counter is not incremented. If the primitive
	 *  does not fit in the buffer object, the counter is not incremented.
	 *  @author Adil Yalcin
	 */
	class RENGAPI GPUQueryPrimTransFB : public GPUQueryPrim {
	public:
		GPUQueryPrimTransFB();

		//! Transform-feedback-primitives-written count maintained by the GL is set to zero.
		void begin();
		void end();
	};

	/**
	 *  @class GPUQueryOcclusion
	 *  @brief Occlusion queries are used
	 *  - to count the number of fragments or samples that pass the depth test, or
	 *  - to detect if any fragments or samples pass the depth test.
	 *  @par
	 *  When an occlusion query is active, the samples-passed count is incremented for each fragment
	 *  that passes the depth test. When a boolean (any) query is made, GL tracks if any samples have
	 *  passed the test.
	 *  @note The two queries summarized above are very similar in functionality and they
	 *        cannot be activated at the same time.
	 *  @author Adil Yalcin
	 */
	class RENGAPI GPUQueryOcclusion : public GPUQuery {
	public:
		GPUQueryOcclusion();

		//! The query object will be used to store the number of fragments or samples that pass
		void begin();
		//! The query object will be used to detect if any fragments or samples pass the test (boolean)
		void beginAny();
		//! Ends the active query (counting or any).
		void end();
	};


	/**
	 *  @class GPUQueryTimer
	 *  @brief Timer queries are used 
	 *  - to track the amount of time needed to fully execute a set of GL commands, or
	 *  - to determine the current time of the GL.
	 *  @note This class holds an abstraction over two different timer queries, 
	 *        you can activate only one type of query for a given query object.
	 *  @author Adil Yalcin
	 */
	class RENGAPI GPUQueryTimer : public GPUQuery {
	public:
		GPUQueryTimer();

		//! Starts the timer
		void begin();
		//! Stops the timer
		void end();

		//! Records the time after all previous commands on the GL client and server state
		//! and the framebuffer have been fully realized.
		//! @note You can access the query result using the generic methods.
		void recordASynchGPUTime();

		//! Returns the synchronous time GPU, that is, when the OpenGL command 
		//! reaches the GL server
		int getSynchGPUTime() const;
		
		//! Returns the synchronous time GPU, that is, when the OpenGL command 
		//! reaches the GL server. (64-bit integer result variant)
		int64 getSynhGPUTime64() const;
	};
}

#endif // RENG_GL_PLATFORM != RENG_GL_PLATFORM_ES

#endif // _RENG_GPUQUERY_H_
